unit uAlgorithms;

interface

uses SysUtils,Classes,Forms,Graphics,Math;

const //Evolution
     PopulationSize = 100;
     NumberOfCycles = 1000000;
     PRepro         = 0.5;
     PMut           = 0.05;
     MagicVal       = 5;
     VectorSize     = MagicVal * MagicVal;
     MagicNumber    = ((VectorSize * (VectorSize+1) div 2)) div MagicVal;

type TVectorValue = array[1..VectorSize] of integer; //zlozky vektorov nebudu binarne ale budu mat celociselne hodnoty

type TVector = class(TPersistent)
       public

       Value:TVectorValue;
       Size:integer;

       constructor Create;
       destructor Destroy;override;

       procedure Generate;
       procedure AssignValue(v:TVectorValue);
       procedure Reset;
       procedure Assign(Source:TPersistent);override;
     end;

type TChromosome = class(TVector)
       public

       szx:integer; //dlzka strany magickeho stvorca (sqrt(ChromozomeSize))

       constructor Create;
       destructor Destroy;override;

       function  Fitness:Double;
       procedure CrossOver(Ch:TChromosome;var Child1,Child2:TChromosome);
       procedure Mutation;
     end;

type TPopulation = class
       public

       Chromosomes:array[1..PopulationSize] of TChromosome;
       NextChromosomes:array[1..PopulationSize] of TChromosome;
       Count:integer;
       Fitnesses:array[1..PopulationSize] of Double;
       Best:TChromosome;

       constructor Create;
       destructor Destroy;override;
       procedure Generate;
       procedure ComputeFitnesses(t:integer);
       function  RouletteWheel:integer;
       procedure Reproduction(idx1,idx2,next:integer);
       procedure Evolution(var AOpt:TChromosome;var Found:boolean);
     end;

const CountsMax = VectorSize*VectorSize;

implementation

uses uMain;

constructor TVector.Create;
begin
  inherited Create;

  Size:=VectorSize;
  FillChar(Value,SizeOf(Value),0);
end;

destructor TVector.Destroy;
begin
  inherited Destroy;
end;

procedure TVector.Generate;
var i,j:integer;
    r:integer;
    OK:boolean;
begin
  for i:=1 to Size do
  begin
    repeat
      r:=random(Size)+1;
      Value[i]:=r;

      if i>1 then
      begin
        for j:=1 to i-1 do
        begin
          OK:=Value[j]<>r;
          if not OK then break;
        end;
      end
      else OK:=true;
    until OK;
  end;
end;

procedure TVector.AssignValue(v:TVectorValue);
begin
  Value:=v;
end;

procedure TVector.Reset;
begin
  Size:=VectorSize;
  FillChar(Value,SizeOf(Value),0);
end;

procedure TVector.Assign(Source: TPersistent);
begin
  if Source is TVector then
  begin
    Value:=TVector(Source).Value;
    Size:=TVector(Source).Size;
  end;
end;

constructor TChromosome.Create;
begin
  inherited Create;

  szx:=MagicVal;
end;

destructor TChromosome.Destroy;
begin
  inherited Destroy;
end;

{procedure TChromosome.CrossOver(Ch:TChromosome;var Child1,Child2:TChromosome);
var CrossPoint:integer;
    i,j:integer;
    Fi:array[1..VectorSize] of integer;   //opravovacia funkcia
    Fi1:array[1..VectorSize] of integer;  //Fi na minus prvu
begin
  FillChar(Fi,Sizeof(Fi),0);
  FillChar(Fi1,Sizeof(Fi1),0);

//  CrossPoint:=Random(szx)*szx+1;
  CrossPoint:=Random(Size)+1;

  if CrossPoint>1 then
  begin
    for i:=1 to CrossPoint-1 do
    begin
      Child1.Value[i]:=Value[i];
      Child2.Value[i]:=Ch.Value[i];
    end;
  end;

  for i:=CrossPoint to Size do
  begin
    Child1.Value[i]:=Ch.Value[i];
    Fi[Value[i]]:=Ch.Value[i];
    Child2.Value[i]:=Value[i];
    Fi1[Ch.Value[i]]:=Value[i];
  end;

  if CrossPoint>1 then
  begin
    for i:=1 to CrossPoint-1 do
    begin
      j:=Child1.Value[i];
      while Fi1[j]<>0 do j:=Fi1[j];
      Child1.Value[i]:=j;

      j:=Child2.Value[i];
      while Fi[j]<>0 do j:=Fi[j];
      Child2.Value[i]:=j;
    end;
  end;
end;}

procedure TChromosome.CrossOver(Ch:TChromosome;var Child1,Child2:TChromosome);
var CrossPoint:integer;
    i,j:integer;
    Fi:array[1..VectorSize] of integer;   //opravovacia funkcia
    Fi1:array[1..VectorSize] of integer;  //Fi na minus prvu
begin
(*  child1.Value:=Value;
  child2.value:=ch.value;
  exit;*)

  FillChar(Fi,Sizeof(Fi),0);
  FillChar(Fi1,Sizeof(Fi1),0);

if random<0.5 then
begin
  CrossPoint:=Random(szx)*szx+1;

  if CrossPoint>1 then
  begin
    for i:=1 to CrossPoint-1 do
    begin
      Child1.Value[i]:=Value[i];
      Child2.Value[i]:=Ch.Value[i];
    end;
  end;

  if CrossPoint+szx<Size then
  begin
    for i:=CrossPoint+szx to Size do
    begin
      Child1.Value[i]:=Value[i];
      Child2.Value[i]:=Ch.Value[i];
    end;
  end;

  for i:=CrossPoint to CrossPoint+szx-1 do
  begin
    Child1.Value[i]:=Ch.Value[i];
    Fi[Value[i]]:=Ch.Value[i];
    Child2.Value[i]:=Value[i];
    Fi1[Ch.Value[i]]:=Value[i];
  end;

  if CrossPoint>1 then
  begin
    for i:=1 to CrossPoint-1 do
    begin
      j:=Child1.Value[i];
      while Fi1[j]<>0 do j:=Fi1[j];
      Child1.Value[i]:=j;

      j:=Child2.Value[i];
      while Fi[j]<>0 do j:=Fi[j];
      Child2.Value[i]:=j;
    end;
  end;

  if CrossPoint+szx<Size then
  begin
    for i:=CrossPoint+szx to Size do
    begin
      j:=Child1.Value[i];
      while Fi1[j]<>0 do j:=Fi1[j];
      Child1.Value[i]:=j;

      j:=Child2.Value[i];
      while Fi[j]<>0 do j:=Fi[j];
      Child2.Value[i]:=j;
    end;
  end;
end
else begin
  CrossPoint:=Random(szx)+1;

  for i:=1 to Size do
  begin
    if i mod CrossPoint<>0 then
    begin
      Child1.Value[i]:=Value[i];
      Child2.Value[i]:=Ch.Value[i];
    end
    else begin
      Child1.Value[i]:=Ch.Value[i];
      Child2.Value[i]:=Value[i];
      Fi[Value[i]]:=Ch.Value[i];
      Fi1[Ch.Value[i]]:=Value[i];
    end;
  end;

  for i:=1 to Size do
  begin
    if i mod CrossPoint<>0 then
    begin
      j:=Child1.Value[i];
      while Fi1[j]<>0 do j:=Fi1[j];
      Child1.Value[i]:=j;

      j:=Child2.Value[i];
      while Fi[j]<>0 do j:=Fi[j];
      Child2.Value[i]:=j;
    end;
  end;
end;
end;

procedure TChromosome.Mutation;
var i,j:integer;
    v:integer;
begin
  for i:=1 to Size do
    if Random<PMut then
    begin
      v:=Value[i];
      if Random<0.5 then inc(Value[i]) else dec(Value[i]);
      if Value[i]=0 then Value[i]:=2;
      if Value[i]=Size+1 then Value[i]:=Size-1;
      for j:=1 to Size do
        if (i<>j) and (Value[j]=Value[i]) then Value[j]:=v;
    end;

{  for i:=1 to Size do
    if Random<PMut then
    begin
      j:=Random(Size)+1;
      v:=Value[i];
      Value[i]:=Value[j];
      Value[j]:=v;
    end;}
end;

{function TChromosome.Fitness:integer;
var x,y:integer;
    XSums:array[1..VectorSize] of integer;  //sem sa ulozia sumy jednotlivych stlpcov v magickom stvorci
    YSums:array[1..VectorSize] of integer;  //to iste s riadkami
    DSums:array[1..2] of integer;           //a s diagonalami
    Counts:array[1..CountsMax] of integer;  //napr. Counts[10] bude obsahovat kolko riadkov, stlpcov a diagonal ma sucet 10
    max:integer;
begin
  FillChar(Counts,SizeOf(Counts),0);
  FillChar(XSums,SizeOf(XSums),0);
  FillChar(YSums,SizeOf(YSums),0);
  DSums[1]:=0;
  DSums[2]:=0;

  for y:=1 to szx do
  begin
    for x:=1 to szx do
    begin
      YSums[x]:=YSums[x]+Value[(y-1)*szx+x];
      XSums[y]:=XSums[y]+Value[(y-1)*szx+x];
      if x=y then DSums[1]:=DSums[1]+Value[(y-1)*szx+x];
      if y=szx-x+1 then DSums[2]:=DSums[2]+Value[(y-1)*szx+x];
    end;
  end;

  for x:=1 to szx do
  begin
    inc(Counts[XSums[x]]);
    inc(Counts[YSums[x]]);
  end;
  inc(Counts[DSums[1]]);
  inc(Counts[DSums[2]]);

  max:=1;
  for x:=1 to CountsMax do
    if Counts[x]>max then max:=Counts[x];
  result:=max;

  result:=max+Counts[MagicNumber];
end;}

{function TChromosome.Fitness:integer;
var x,y:integer;
    XSums:array[1..VectorSize] of integer;  //sem sa ulozia sumy jednotlivych stlpcov v magickom stvorci
    YSums:array[1..VectorSize] of integer;  //to iste s riadkami
    DSums:array[1..2] of integer;           //a s diagonalami
    Counts:array[1..CountsMax] of integer;  //napr. Counts[10] bude obsahovat kolko riadkov, stlpcov a diagonal ma sucet 10
    max,res:integer;
begin
  FillChar(Counts,SizeOf(Counts),0);
  FillChar(XSums,SizeOf(XSums),0);
  FillChar(YSums,SizeOf(YSums),0);
  DSums[1]:=0;
  DSums[2]:=0;

  for y:=1 to szx do
  begin
    for x:=1 to szx do
    begin
      YSums[x]:=YSums[x]+Value[(y-1)*szx+x];
      XSums[y]:=XSums[y]+Value[(y-1)*szx+x];
      if x=y then DSums[1]:=DSums[1]+Value[(y-1)*szx+x];
      if y=szx-x+1 then DSums[2]:=DSums[2]+Value[(y-1)*szx+x];
    end;
  end;

  for x:=1 to szx do
  begin
    inc(Counts[XSums[x]]);
    inc(Counts[YSums[x]]);
  end;

  res:=0;
  if DSums[1]=MagicNumber then inc(res,szx);
  if DSums[2]=MagicNumber then inc(res,szx);

  result:=res+Counts[MagicNumber];
end;}

function TChromosome.Fitness:Double;
var x,y:integer;
    XSums:array[1..VectorSize] of integer;  //sem sa ulozia sumy jednotlivych stlpcov v magickom stvorci
    YSums:array[1..VectorSize] of integer;  //to iste s riadkami
    DSums:array[1..2] of integer;           //a s diagonalami
//    Counts:array[1..CountsMax] of integer;  //napr. Counts[10] bude obsahovat kolko riadkov, stlpcov a diagonal ma sucet 10
    res:Double;
begin
//  FillChar(Counts,SizeOf(Counts),0);
  FillChar(XSums,SizeOf(XSums),0);
  FillChar(YSums,SizeOf(YSums),0);
  DSums[1]:=0;
  DSums[2]:=0;

  for y:=1 to szx do
  begin
    for x:=1 to szx do
    begin
      YSums[x]:=YSums[x]+Value[(y-1)*szx+x];
      XSums[y]:=XSums[y]+Value[(y-1)*szx+x];
      if x=y then DSums[1]:=DSums[1]+Value[(y-1)*szx+x];
      if y=szx-x+1 then DSums[2]:=DSums[2]+Value[(y-1)*szx+x];
    end;
  end;

  res:=1;
  for x:=1 to szx do
  begin
    res:=res+abs(XSums[x]-MagicNumber);
    res:=res+abs(YSums[x]-MagicNumber);
  end;
  res:=res+abs(DSums[1]-MagicNumber);
  res:=res+abs(DSums[2]-MagicNumber);

  result:=1/res;
end;

constructor TPopulation.Create;
var i:integer;
begin
  inherited Create;

  Count:=PopulationSize;
  for i:=1 to Count do
  begin
    Chromosomes[i]:=TChromosome.Create;
    NextChromosomes[i]:=TChromosome.Create;
  end;

  Best:=TChromosome.Create;
end;

destructor TPopulation.Destroy;
var i:integer;
begin
  for i:=1 to Count do
  begin
    Chromosomes[i].Free;
    NextChromosomes[i].Free;
  end;

  Best.Free;

  inherited Destroy;
end;

procedure TPopulation.Generate;
var i:integer;
begin
  for i:=1 to Count do
  begin
    Chromosomes[i].Generate;;
  end;

  Best.Assign(Chromosomes[1]);
end;

procedure TPopulation.ComputeFitnesses(t:integer);
var i,x,y:integer;
    MaxFit:Double;
    LocalBest:TChromosome;
    PSum:Double;
    s:string;
    SumFit:Double;
begin
  SumFit:=0;
  MaxFit:=0;
  for i:=1 to Count do
  begin
    Fitnesses[i]:=Chromosomes[i].Fitness;
    SumFit:=SumFit+Fitnesses[i];

    if MaxFit<Fitnesses[i] then
    begin
      LocalBest:=Chromosomes[i];
      MaxFit:=Fitnesses[i];
    end;
  end;

  frmMain.Canvas.Pixels[t mod 600,500-Round(MaxFit*100)]:=clOlive;

  if Best.Fitness<MaxFit then
  begin
    Best.Assign(LocalBest);
    s:='';
    for y:=1 to LocalBest.szx do
    begin
      s:='';
      for x:=1 to LocalBest.szx do
      begin
        s:=s+IntToStr(LocalBest.Value[(y-1)*LocalBest.szx+x])+' ';
      end;
      frmMain.lbOut.Items.Add(s);
    end;
    frmMain.lbOut.Items.Add('  ');
    Application.ProcessMessages;
  end;

  PSum:=0;
  for i:=1 to Count do
  begin
    PSum:=PSum+Fitnesses[i];
    Fitnesses[i]:=Round(PSum/SumFit*2147483648);
  end;
end;

function TPopulation.RouletteWheel:integer;
var i,r,mini,maxi:integer;
    done:boolean;
begin
  Done:=false;
  MinI:=1;
  MaxI:=Count;

  r:=random(2147483648);

  repeat
    i:=(MinI+MaxI) shr 1;

    if r<Fitnesses[i] then
    begin
      if (i=1) or (r>=Fitnesses[i-1]) then
        Done:=true
      else
        MaxI:=i;
    end
    else begin
      MinI:=i+1;
    end;
  until Done;

  result:=i;
end;

procedure TPopulation.Reproduction(idx1,idx2,next:integer);
begin
  Chromosomes[idx1].CrossOver(Chromosomes[idx2],NextChromosomes[next],NextChromosomes[next+1]);
  NextChromosomes[next].Mutation;
  NextChromosomes[next+1].Mutation;
end;

procedure TPopulation.Evolution(var AOpt:TChromosome;var Found:boolean);
var i,t:integer;
    StopCriterion:boolean;
    NextPopulationSize:integer;
    idx1,idx2:integer;
begin
  Found:=false;

  Generate;

  t:=0;
  while (t<NumberOfCycles) and (not Found) do
  begin
    t:=t+1;

    ComputeFitnesses(t);

    if Best.Fitness=1 then
    begin
      Beep;
      Found:=true;
      break;
    end;

    NextChromosomes[1].AssignValue(Best.Value);
    NextChromosomes[2].AssignValue(Best.Value);

    NextPopulationSize:=2;
    while NextPopulationSize<Count do
    begin
      idx1:=RouletteWheel;
      idx2:=RouletteWheel;
      if idx1=idx2 then Continue;

      if Random<PRepro then
      begin
        Reproduction(idx1,idx2,NextPopulationSize+1);
      end
      else begin
        NextChromosomes[NextPopulationSize+1].AssignValue(Chromosomes[idx1].Value);
        NextChromosomes[NextPopulationSize+2].AssignValue(Chromosomes[idx2].Value);
      end;

      inc(NextPopulationSize,2);
    end;

//    Count:=NextPopulationSize;
    for i:=1 to Count do
      Chromosomes[i].AssignValue(NextChromosomes[i].Value);

    frmMain.lblGeneration.Caption:=IntToStr(t);

    frmMain.Canvas.Pixels[t mod 600,500-Round(Best.Fitness*100)]:=clBlack;
    frmMain.Canvas.Pixels[t mod 600,500]:=clRed;
    frmMain.Canvas.Pixels[t mod 600,500-100]:=clGreen;

    Application.ProcessMessages;

    if StopIt then break;
  end;

  AOpt:=TChromosome.Create;
  AOpt.Assign(Best);
end;

end.
